/*
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%        one step t --> t+h of the Gear Predictor-Corrector method           %
%                 for second order differential equations                    %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

 d2a/dt2 = b(t,a,da/dt) (#ifdef VELOCITY)

   or

 d2a/dt2 = b(t,a) (#ifndef VELOCITY)

   is solved, where a,b are vectors of neq real elements and the right
   hand side, b(t,a[,da/dt]), is computed by procedure rhs(b,a[,da/dt])

 Globals required:
   real t=time, h=time step
   int GearOrder = order in {3,4,5,6}
 typedef required:
   ToIntPtr=pointer to anything containing a contiguous vector of neq reals
 Macro or function giving pointer to the first real of the vector:
   D(a)
 If VELOCITY is #defined, rhs is assumed to depend on da/dt

 The contents of  a  and  da/dt  should not be changed except small
 corrections of the order of the method (e.g., to correct constraints).

 typedef real must be float or double
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
see gear.c for equation da/dt = b(t,a)
*/

#ifdef PAR
#  include "cputime.h"
#endif /*# PAR */


#  define MAXGEARORDER 8

static real GearCoef[MAXGEARORDER-2][MAXGEARORDER] = {
  /*3*/ {0.0,        1.0,         1.0},
  /*4*/ {1.0/6,      5.0/6,       1.0, 1.0/3},
  /*5*/ {19.0/120,   0.75,        1.0, 0.5,     1.0/12},
  /*6*/ {3.0/20,     251.0/360,   1.0, 11.0/18, 1.0/6,  1.0/60},
  /*7*/ {863./6048,  665./1008,   1.0, 25./36,  35./144,1./24,  1./360},
  /*8*/ {1925./14112,19087./30240,1.0,137./180, 5./16,  17./240,1./120,1./2520} };

static real GearCoefv[MAXGEARORDER-2][MAXGEARORDER] = {
  /*3*/ {0.0,        1.0,         1.0},
  /*4*/ {1.0/6,      5.0/6,       1.0, 1.0/3},
  /*5*/ {19.0/90,    0.75,        1.0, 0.5,     1.0/12},
  /*6*/ {3.0/16,     251.0/360,   1.0, 11.0/18, 1.0/6,  1.0/60},
  {-9e9},{-9e9} };


void Gear(int neq,ToIntPtr *a) /*************************************** Gear */
{
  int i,j,k,bin;
  real hh2=h*h/2,s,d,ss,*r,*hdr,*pred,*rhspred;
  static real *gc=NULL;
#  ifdef VELOCITY
  real *preddr;
#  endif /*# VELOCITY */

  if (GearOrder<3 || GearOrder>MAXGEARORDER) ERROR(("bad GearOrder (option -m)"))
  if (gear.lastorder && GearOrder!=gear.lastorder) ERROR(("change of GEARORDER (option -m) not supported"))
  gear.lastorder=GearOrder;

  r=D(a[0]);
  hdr=D(a[2]);
  pred=D(a[GearOrder]);
#  ifdef VELOCITY
  preddr=D(a[GearOrder+2]);
#  endif /*# VELOCITY */
  rhspred=D(a[GearOrder+1]);

  if (gear.init<0) {
    /* old partly optimized code */

    if (gear.init==-1) gc=GearCoefv[GearOrder-3];
    else gc=GearCoef[GearOrder-3];
    if (gc[0]<1e-5) ERROR(("gear2: coefficients not known (use gear.init=-2 or lower order))"))

    switch (GearOrder) {

      case 4: /* optimized code */

        /* predictor for r and dr/dt */
        loop (i,0,neq) {
          pred[i] = (D(a[3]))[i] + (D(a[2]))[i] + (D(a[1]))[i] + (D(a[0]))[i];
#  ifdef VELOCITY
          preddr[i] = (3*(D(a[3]))[i] + 2*(D(a[2]))[i] + (D(a[1]))[i])/h;
#  endif /*# VELOCITY */
        }

        /* r.h.s. */
        t += h;

#  ifdef PAR
                                                              CPUtime("gear");
#  endif /*# PAR */

#  ifdef VELOCITY
        rhs(a[5],a[4],a[6]);
#  else /*# VELOCITY */
        rhs(a[5],a[4]);
#  endif /*#!VELOCITY */

        /* corrector and the rest of predictor */
        loop (i,0,neq) {
          ss = hh2*rhspred[i];
          d = ss-hdr[i]-3*(D(a[3]))[i];
          r[i] = pred[i] + d*gc[0];
#  ifdef VELOCITY
          (D(a[1]))[i] = h*preddr[i] + d*gc[1];
#  else /*# VELOCITY */
          (D(a[1]))[i] = d*gc[1] + 3*(D(a[3]))[i] + 2*(D(a[2]))[i] + (D(a[1]))[i];
#  endif /*#!VELOCITY */
          (D(a[3]))[i] = d*gc[3] + (D(a[3]))[i];
          hdr[i] = ss; }

        break;

      default: /* general case - slow code */

        /* predictor */
        loop (i,0,neq) {
          s = 0;
          loop (j,0,GearOrder) s += (D(a[j]))[i];
          pred[i] = s;
#  ifdef VELOCITY
          s = 0;
          loop (j,1,GearOrder) s += j*(D(a[j]))[i];
          preddr[i] = s/h;
#  endif /*# VELOCITY */
        }

        /* r.h.s. */
        t += h;

#  ifdef VELOCITY
        rhs(a[GearOrder+1],a[GearOrder],a[GearOrder+2]);
#  else /*# VELOCITY */
        rhs(a[GearOrder+1],a[GearOrder]);
#  endif /*#!VELOCITY */

        /* corrector and the rest of predictor */
        loop (i,0,neq) {
          ss = hh2*rhspred[i];
          d = ss-hdr[i];
          loop (j,3,GearOrder) d -= j*(j-1)/2*(D(a[j]))[i];
          r[i] = pred[i] + d*gc[0];
#  ifdef VELOCITY
          (D(a[1]))[i] = h*preddr[i] + d*gc[1];
#  else /*# VELOCITY */
          s = d*gc[1];
          loop (j,1,GearOrder) s += j*(D(a[j]))[i];
          (D(a[1]))[i]=s;
#  endif /*#!VELOCITY */
          loop (j,3,GearOrder) {
            s = d*gc[j]; bin = 1;
            loop (k,j,GearOrder) {
              s += (D(a[k]))[i]*bin;
              bin = bin*(k+1)/(k+1-j); }
            (D(a[j]))[i] = s; }
          hdr[i] = ss; }

    } /* switch */
    //    prt("%g %g %g %g %g GEAR",(D(a[0]))[99],(D(a[1]))[99],(D(a[2]))[99],(D(a[3]))[99],(D(a[4]))[99]);

  } else {

    /* development version VELOCITY only */
#  ifndef VELOCITY
    ERROR(("not VELOCITY not supported with gear.init>=0"))
#  endif /*# VELOCITY */

#  undef MAXGEARORDER
#  define MAXGEARORDER 8
/* pascal 7 | transtab | repl "  " " " | repl " " , */
    static double PASCAL[MAXGEARORDER][MAXGEARORDER]={
      {1,1,1,1,1,1,1,1},
      {0,1,2,3,4,5,6,7},
      {0,0,1,3,6,10,15,21},
      {0,0,0,1,4,10,20,35},
      {0,0,0,0,1,5,15,35},
      {0,0,0,0,0,1,6,21},
      {0,0,0,0,0,0,1,7},
      {0,0,0,0,0,0,0,1}};
    static double PRED[MAXGEARORDER][MAXGEARORDER];

    if (gear.init>0) {
      if (gear.init<3 || gear.init>MAXGEARORDER) ERROR(("gear.init=%d (Gear's order) out of bounds",gear.init))
      gc=GearCoef[gear.init-3]; /* only no-velocity version */
      prt("\nGear initialized, order=%d\nPredictor matrix:",GearOrder);
      loop (i,0,GearOrder) {
        loop (j,0,GearOrder) {
          PRED[i][j]=PASCAL[i][j]*gear.P[j];
          prt_("%12.8f",PRED[i][j]); }
        _n
        if (gear.C[i]>-1e5) gc[i]=gear.C[i]; }
      prt("Corector coefficients:");
      loop (j,0,GearOrder) prt_("%12.8f",gc[j]);
      _n _n
      gear.init=0; }

    if (!gc) ERROR(("gear2: not initialized, set gear.init"))

    loop (i,0,neq) {
      s = 0;
      loop (j,0,GearOrder) s += PRED[0][j]*(D(a[j]))[i];
      pred[i] = s;
      s = 0;
      loop (j,1,GearOrder) s += PRED[1][j]*(D(a[j]))[i];
      preddr[i] = s/h;
    }

    /* r.h.s. */
    t += h;

    rhs(a[GearOrder+1],a[GearOrder],a[GearOrder+2]);

    /* corrector and the rest of predictor */
    loop (i,0,neq) {
      ss = hh2*rhspred[i];
      d = ss-hdr[i];
      loop (j,3,GearOrder) d -= PRED[2][j]*(D(a[j]))[i];
      r[i] = pred[i] + d*gc[0];
      (D(a[1]))[i] = h*preddr[i] + d*gc[1];
      loop (j,3,GearOrder) {
        s = d*gc[j];
        loop (k,j,GearOrder) {
          s += PRED[j][k]*(D(a[k]))[i];}
        (D(a[j]))[i] = s; }
      hdr[i] = ss; }
    //    prt("%g %g %g %g %g GEAR",(D(a[0]))[99],(D(a[1]))[99],(D(a[2]))[99],(D(a[3]))[99],(D(a[4]))[99]);
  }    
#  ifdef PAR
                                                             CPUtime("gear");
#  endif /*# PAR */
}
